/*
 *  A simple realchat client
 *
 *  Copyright (C) 2005  Andreas Beck <becka-rcc@bedatec.de>
 *
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License version 2 as
 *  published by the Free Software Foundation;
 *
 *  Software distributed under the License is distributed on an "AS
 *  IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 *  implied. See the License for the specific language governing
 *  rights and limitations under the License.
 *
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include <unistd.h>
#include <fcntl.h>
#include <netdb.h>

#include <time.h>

#include <sys/types.h>
#include <sys/socket.h>

char *nickname="Wurstbrot";
char *servername="your.server.goes.here";
int serverport=8080;

enum globstate { RS_NOTHING, RS_GOT_INIT };

enum globstate globstate=RS_NOTHING;

enum runstate { RS_INIT, RS_WAITFORLEN };

void process_inboundchar(unsigned char newchr) {

  static enum runstate runstate=RS_WAITFORLEN;
  static unsigned char buffer[1024];
  static int bufpos=0;
  static int targetlen=4;
  buffer[bufpos++]=newchr;
  if (bufpos==targetlen) {
    switch(buffer[0]) {
      case 0xac: 
        if (bufpos!=4) {
          fprintf(stderr,"ac out of init sequence?\n");
        } else {
          fprintf(stderr,"Got init sequence.\n");
          globstate=RS_GOT_INIT;
        }
        restart_from_0:
          bufpos=0;
          targetlen=1;
          return;
      case 0x71:
        if (bufpos==1) {
          fprintf(stderr,"Got 71.");
          targetlen=5;
        } else if (bufpos==5) {
          if (buffer[4]==0x03)
            targetlen+=4;
          else if (buffer[4]==0x0b)
            goto restart_from_0;
          else if (buffer[4]==0x00)
            targetlen=18;
          else if (buffer[4]==0x01)
            targetlen=16;
          else /* 0x15 is catched by that. */
            goto restart_from_0;
        } else {
          goto restart_from_0;
        }
        break;
      case 0x73:
        if (bufpos!=1) {
          fprintf(stderr,"0x73, but pos!=1?\n");
        } else {
          fprintf(stderr,"NOP (0x73)\n");
        }
        goto restart_from_0;
      case 0x72:
      case 0x74:
      case 0x49:
      case 0x4a:
      case 0x5b:
        if (bufpos==1) {
          targetlen=3;	/* get length */
        } else if (bufpos==3) {
          targetlen=3+buffer[1]*256+buffer[2];
        } else {
          buffer[bufpos]='\0';
          fprintf(stderr,"Got string (id=0x%02x):%s\n",buffer[0],buffer+3);
          goto restart_from_0;
        }
        break;
      case 0xd3:
        if (bufpos==1) {
          targetlen=11;
        } else {
          fprintf(stderr,"Got 0xd3-stuff.11 chars.\n");
          goto restart_from_0;
        }
        break;
      case 0xd9:
        if (bufpos==1) {
          targetlen=11;
        } else {
          fprintf(stderr,"Got 0xd9-stuff.11 chars.\n");
          goto restart_from_0;
        }
        break;
      case 0x78:
        if (bufpos==1) {
          targetlen=15;
        } else {
          fprintf(stderr,"Got 0x78-stuff.15 chars.\n");
          goto restart_from_0;
        }
        break;
      case 0x90:
        if (bufpos==1) {
          targetlen=17;
        } else {
          fprintf(stderr,"Got 0x90-stuff.17 chars.\n");
          goto restart_from_0;
        }
        break;
      default:
        fprintf(stderr,"Got junk 0x%02x.\n",buffer[0]);
        goto restart_from_0;
    }
  }
}

void say_hello(int sock) {

  send(sock,"\xac\xed\x00\x05",4,0);
}

void say_7371(int sock,unsigned char byte9, unsigned char byte17) {

  unsigned char basestring[]=
    "\x73\x71\x00\x7e\x00\x00\x00\x00"
    "\x00\xc9\x00\x00\x00\x00\x00\x00"
    "\x00\x35";

  basestring[ 9]=byte9;
  basestring[17]=byte17;

  send(sock,basestring,18,0);
}

void saystring_num_count(int sock,unsigned char num,unsigned char *string) {

  unsigned char basestring[]="\x74\x00\x00";
  int len;
  
  basestring[0]=num;
  len=strlen(string);

  basestring[1]=len>>8;
  basestring[2]=len&0xff;
  send(sock,basestring,3,0);
  send(sock,string,len,0);
}

void saystring_74(int sock,unsigned char *string) {

  saystring_num_count(sock,0x74,string);
}

void saystring_72(int sock,unsigned char *string) {

  saystring_num_count(sock,0x72,string);
}

void saystring_verbatim(int sock,unsigned char *string,int len) {

  send(sock,string,len,0);
}

void say_chatline(int sock,unsigned char *string,int namecol) {

  char namecolstring[3]="\x60n";

  say_7371(sock,0x69,0x1e);
  saystring_verbatim(sock,"\x75\x71\x00\x7e\x00\x03\x00\x00\x00\x02",10);
  namecolstring[1]=namecol;
  saystring_74(sock,namecolstring);
  saystring_74(sock,string);
}

void say_71007e000b(int sock) {
  saystring_verbatim(sock,"\x71\x00\x7e\x00\x0b",5);
}

void say_keepalive(int sock) {
  say_7371(sock,0x66,0x01);
  saystring_verbatim(sock,"\x70",1);
}

void say_login(int sock,unsigned char *hostname,int avatar) {
  unsigned char buffer[1024];
  sprintf(buffer,"HOST: %s; 3.5.1b; Java: 1.5.0_01; Sun Microsystems Inc.; Linux; null; %d",
    hostname,avatar);
  saystring_74(sock,buffer); 
}

int open_direct_socket(char *servername,int port) {

  int sock;
  int proto_tcp;
  struct protoent *pe;
  struct hostent *he;
  struct sockaddr_in sa;

  pe=getprotobyname("tcp");
  proto_tcp=pe->p_proto;
  sock=socket(PF_INET,SOCK_STREAM,proto_tcp);

  he=gethostbyname(servername);

  sa.sin_family=AF_INET;
  sa.sin_port=htons(port);
  memcpy(&sa.sin_addr.s_addr,he->h_addr,sizeof(sa.sin_addr.s_addr));
  connect(sock,(struct sockaddr *)&sa,sizeof(sa));

  return sock;
}

int main(int argc, char *argv[]) {

  int sock;
  unsigned char bla;
  unsigned char *satz=NULL;
  unsigned char sendbuf[1024];
  int sendbuflen;
  int avatar=4;
  time_t lastka;
  int namecol='n';

  sendbuflen=0;

  if (argc>=2) nickname=argv[1];
  if (argc>=3) avatar=atoi(argv[2]);
  if (argc>=4) satz=argv[3];
  
  sock=open_direct_socket(servername,serverport);

  say_hello(sock);
  saystring_verbatim(sock,"\x73",1);
  saystring_72(sock,"rcs.comm.cMsg");
  saystring_verbatim(sock,"\xd3\x94\x62\x76\xba\xd5\xbf\x62\x02\x00\x03",11);
  saystring_num_count(sock,0x49,"commandID");
  saystring_num_count(sock,0x4A,"size");
  saystring_num_count(sock,0x5b,"arg");
  saystring_74(sock,"[Ljava/lang/Object;");
  saystring_verbatim(sock,"\x78\x70\x00\x00\x00\x67\x00\x00\x00\x00\x00\x00\x00\x5b\x75",15);
  saystring_72(sock,"[Ljava.lang.Object;");
  saystring_verbatim(sock,"\x90\xce\x58\x9f\x10\x73\x29\x6c\x02\x00\x00\x78",12);
  saystring_verbatim(sock,"\x70\x00\x00\x00\x04",5);
  saystring_74(sock,nickname);	/* Send the Nick */
  saystring_verbatim(sock,"\x70\x70",2);
  //saystring_74(sock,"HOST: www.jappy.de; 3.5.1b; Java: 1.5.0_01; Sun Microsystems Inc.; Linux; null; 4");
  say_login(sock,"www.jappy.de",avatar);

  fprintf(stderr,"Sent logon stuff.\n");

  while(globstate!=RS_GOT_INIT) {
    if(recv(sock,&bla,1,MSG_DONTWAIT)==1) {
      process_inboundchar(bla);
    }
  }

  say_7371(sock,0x68,0x01);
  saystring_verbatim(sock,"\x75\x71\x00\x7e\x00\x03\x00\x00\x00\x01\x75",11);
  saystring_72(sock,"[Ljava.lang.String;");
  saystring_verbatim(sock,"\xad\xd2\x56\xe7\xe9\x1d\x7b\x47\x02\x00\x00\x78\x70",13);
  saystring_verbatim(sock,"\x00\x00\x00\x08\x74\x00\x00",7);
  say_71007e000b(sock);
  say_71007e000b(sock);
  say_71007e000b(sock);
  say_71007e000b(sock);
  say_71007e000b(sock);
  say_71007e000b(sock);
  say_71007e000b(sock);
  while(recv(sock,&bla,1,MSG_DONTWAIT)==1) {
    // fprintf(stderr,"%02x",bla);
    process_inboundchar(bla);
  }
  fprintf(stderr,"Sent secondary stuff.\n");

  sleep(1);
  while(recv(sock,&bla,1,MSG_DONTWAIT)==1) {
    // fprintf(stderr,"%02x",bla);
    process_inboundchar(bla);
  }
  if (satz) say_chatline(sock,satz,'n');

  fprintf(stderr,"Sent chatline stuff.\n");

  fcntl(0,F_SETFL,O_NONBLOCK);
  lastka=time(NULL);
  while(1) {
    if (recv(sock,&bla,1,MSG_DONTWAIT)==1) {
      // fprintf(stderr,"%02x",bla);
      process_inboundchar(bla);
    } else if (read(0,sendbuf+sendbuflen,1)==1) {
      if (sendbuf[sendbuflen]=='\r'||sendbuf[sendbuflen]=='\n') {
        sendbuf[sendbuflen]='\0';
        if (sendbuf[0]=='/') {
          if (strncmp(sendbuf,"/col",4)==0) {
            namecol=sendbuf[5];
          } else {
            fprintf(stderr,"Unknown command.\n");
          }
        } else {
          say_chatline(sock,sendbuf,namecol);
        }
        sendbuflen=0;
      } else {
        sendbuflen++;
      }
    } else sleep(1);
    if (time(NULL)-lastka>5) {
      say_keepalive(sock);
      lastka=time(NULL);
    }
  }
}
